@@ -1,9 +1,13 @@
 use super::*;
 
+use std::mem;
+
 #[test]
 fn tuples_have_correct_size() {
-    assert_eq!(12, (1u16, 2u32, 3i16, 'x').mem_size());
-    assert_eq!(5, ((1u8, 2i8, 3u8), 4i16).mem_size());
+    assert_eq!(mem::size_of::<(u16, u32, i16, char)>(),
+        (1u16, 2u32, 3i16, 'x').mem_size());
+    assert_eq!(mem::size_of::<((u8, i8, u8), i16)>(),
+        ((1u8, 2i8, 3u8), 4i16).mem_size());
 }
 
 #[test]
@@ -40,7 +44,7 @@ fn options_have_correct_size() {
 
 #[test]
 fn entry_correctly_computes_size() {
-    let entry = Entry::new("hello".to_owned(), "world!".to_owned());
+    let entry = UnhingedEntry::new("hello".to_owned(), "world!".to_owned());
 
     let key_str_bytes = 5;
     let value_str_bytes = 6;
@@ -58,7 +62,7 @@ fn entry_correctly_computes_size() {
         + usize_bytes
         + 2 * ptr_bytes;
 
-    assert_eq!(expected_bytes, entry.size);
+    assert_eq!(expected_bytes, entry.size());
 }
 
 #[test]
@@ -136,8 +140,8 @@ fn cache_ejects_lru_if_overflowing() {
     assert!(cache.get("b").is_some());
     assert!(cache.get("c").is_some());
 
-    assert_eq!("b", unsafe { &(*cache.tail).key });
-    assert_eq!("c", unsafe { &(*cache.head).key });
+    assert_eq!("b", unsafe { cache.seal.get().prev.get().key() });
+    assert_eq!("c", unsafe { cache.seal.get().next.get().key() });
 }
 
 #[test]
@@ -156,13 +160,13 @@ fn getting_sets_most_recently_used() {
     assert!(cache.get("b").is_none());
     assert!(cache.get("c").is_some());
 
-    assert_eq!("a", unsafe { &(*cache.tail).key });
-    assert_eq!("c", unsafe { &(*cache.head).key });
+    assert_eq!("a", unsafe { cache.seal.get().prev.get().key() });
+    assert_eq!("c", unsafe { cache.seal.get().next.get().key() });
 
     cache.get("a");
 
-    assert_eq!("c", unsafe { &(*cache.tail).key });
-    assert_eq!("a", unsafe { &(*cache.head).key });
+    assert_eq!("c", unsafe { cache.seal.get().prev.get().key() });
+    assert_eq!("a", unsafe { cache.seal.get().next.get().key() });
 }
 
 #[test]
@@ -189,7 +193,7 @@ fn cache_rejects_too_large_entry() {
         entry over the capacity of the cache.".to_owned();
 
     assert!(matches!(cache.insert(key, value),
-        Err(LruError::EntryTooLarge { .. })));
+        Err(InsertError::EntryTooLarge { .. })));
 }
 
 #[test]
@@ -214,6 +218,53 @@ fn precisely_fitting_entry_does_not_eject_lru() {
     assert_eq!(3, cache.len());
 }
 
+#[test]
+fn try_insert_works_as_insert_if_ok() {
+    let mut cache = LruCache::new(1024);
+    cache.insert("hello".to_owned(), "world".to_owned()).unwrap();
+    cache.try_insert("greetings".to_owned(), "moon".to_owned()).unwrap();
+
+    assert_eq!(2, cache.len());
+    assert_eq!("world".to_owned(), cache.remove_lru().unwrap().1);
+    assert_eq!("moon".to_owned(), cache.remove_lru().unwrap().1);
+}
+
+#[test]
+fn try_insert_fails_on_duplication() {
+    let mut cache = LruCache::new(1024);
+    cache.insert("hello".to_owned(), "world".to_owned()).unwrap();
+    let result = cache.try_insert("hello".to_owned(), "moon".to_owned());
+
+    assert!(matches!(result, Err(TryInsertError::OccupiedEntry { .. })));
+    assert_eq!(1, cache.len());
+    assert_eq!(&"world".to_owned(), cache.get("hello").unwrap());
+}
+
+#[test]
+fn try_insert_fails_if_eject_required() {
+    let mut cache = LruCache::new(1024);
+    cache.insert("hello".to_owned(), "world".to_owned()).unwrap();
+    cache.insert("greetings".to_owned(), "moon".to_owned()).unwrap();
+    cache.set_max_size(cache.current_size());
+    let result = cache.try_insert("ahoy".to_owned(), "mars".to_owned());
+
+    assert!(matches!(result, Err(TryInsertError::WouldEjectLru { .. })));
+    assert_eq!(2, cache.len());
+    assert!(!cache.contains("ahoy"));
+}
+
+#[test]
+fn try_insert_fails_if_too_large() {
+    let mut cache = LruCache::new(1024);
+    cache.insert("hello".to_owned(), "world".to_owned()).unwrap();
+    let value = String::from_utf8(vec![b'0'; 1024]).unwrap();
+    let result = cache.try_insert("key".to_owned(), value);
+
+    assert!(matches!(result, Err(TryInsertError::EntryTooLarge { .. })));
+    assert_eq!(1, cache.len());
+    assert!(!cache.contains("key"));
+}
+
 #[test]
 fn removing_works() {
     let mut cache = LruCache::new(1024);
@@ -225,6 +276,56 @@ fn removing_works() {
     assert_eq!(None, cache.remove("hello"));
 }
 
+#[test]
+fn retain_does_not_affect_empty_cache() {
+    let mut cache: LruCache<u64, Vec<u8>> = LruCache::new(1024);
+    cache.retain(|k, v| v.len() as u64 == *k);
+
+    assert_eq!(0, cache.len());
+}
+
+#[test]
+fn retain_works_if_all_match() {
+    let mut cache: LruCache<u64, Vec<u8>> = LruCache::new(1024);
+    cache.insert(1, vec![0]).unwrap();
+    cache.insert(5, vec![2, 3, 5, 7, 11]).unwrap();
+    cache.insert(4, vec![1, 4, 9, 16]).unwrap();
+    cache.retain(|k, v| v.len() as u64 == *k);
+
+    assert_eq!(3, cache.len());
+    assert!(cache.contains(&1));
+    assert!(cache.contains(&4));
+    assert!(cache.contains(&5));
+}
+
+#[test]
+fn retain_works_if_none_match() {
+    let mut cache: LruCache<u64, Vec<u8>> = LruCache::new(1024);
+    cache.insert(1, vec![0, 1]).unwrap();
+    cache.insert(5, vec![2, 3, 5, 7, 11, 13]).unwrap();
+    cache.insert(4, vec![1, 4, 9, 16, 25]).unwrap();
+    cache.retain(|k, v| v.len() as u64 == *k);
+
+    assert_eq!(0, cache.len());
+    assert!(!cache.contains(&1));
+    assert!(!cache.contains(&4));
+    assert!(!cache.contains(&5));
+}
+
+#[test]
+fn retain_works_if_some_match() {
+    let mut cache: LruCache<u64, Vec<u8>> = LruCache::new(1024);
+    cache.insert(1, vec![0, 1]).unwrap();
+    cache.insert(5, vec![2, 3, 5, 7, 11]).unwrap();
+    cache.insert(4, vec![1, 4, 9, 16]).unwrap();
+    cache.retain(|k, v| v.len() as u64 == *k);
+
+    assert_eq!(2, cache.len());
+    assert!(!cache.contains(&1));
+    assert!(cache.contains(&4));
+    assert!(cache.contains(&5));
+}
+
 #[test]
 fn contains_works() {
     let mut cache = LruCache::new(1024);
@@ -288,7 +389,7 @@ fn cache_rejects_too_expanding_mutation() {
         v.append(&mut vec![0u8; 1000]);
     });
 
-    assert!(matches!(result, Err(LruError::EntryTooLarge { .. })));
+    assert!(matches!(result, Err(MutateError::EntryTooLarge { .. })));
     assert_eq!(1, cache.len());
     assert!(cache.current_size() < old_size);
     assert_eq!(None, cache.get("hello"));
@@ -302,7 +403,7 @@ where
     cache.insert(0, 0).unwrap();
     let entry_size = cache.current_size();
 
-    for i in 1..=5000 {
+    for i in 1..=1000 {
         cache.insert(i, i).unwrap();
 
         for j in 0..=(i / 2) {
@@ -368,14 +469,19 @@ fn reserving_fails_on_overflow() {
     assert!(cache.try_reserve(usize::MAX).is_err());
 }
 
-#[test]
-fn shrinking_does_not_drop_below_len() {
+fn large_test_cache() -> LruCache<&'static str, &'static str> {
     let mut cache = LruCache::new(1024);
     cache.insert("hello", "world").unwrap();
     cache.insert("greetings", "moon").unwrap();
     cache.insert("ahoy", "mars").unwrap();
     cache.insert("hi", "venus").unwrap();
     cache.insert("good morning", "jupiter").unwrap();
+    cache
+}
+
+#[test]
+fn shrinking_does_not_drop_below_len() {
+    let mut cache = large_test_cache();
     cache.shrink_to(4);
 
     assert!(cache.capacity() >= 5);
@@ -408,13 +514,7 @@ fn cache_works_with_many_reallocations() {
 
 #[test]
 fn iter_works() {
-    let mut cache = LruCache::new(1024);
-    cache.insert("hello", "world").unwrap();
-    cache.insert("greetings", "moon").unwrap();
-    cache.insert("ahoy", "mars").unwrap();
-    cache.insert("hi", "venus").unwrap();
-    cache.insert("good morning", "jupiter").unwrap();
-
+    let cache = large_test_cache();
     let mut iter = cache.iter();
 
     assert_eq!(Some((&"hello", &"world")), iter.next());
@@ -428,13 +528,7 @@ fn iter_works() {
 
 #[test]
 fn separated_iters_zip_to_pair_iter() {
-    let mut cache = LruCache::new(1024);
-    cache.insert("hello", "world").unwrap();
-    cache.insert("greetings", "moon").unwrap();
-    cache.insert("ahoy", "mars").unwrap();
-    cache.insert("hi", "venus").unwrap();
-    cache.insert("good morning", "jupiter").unwrap();
-
+    let cache = large_test_cache();
     let pair_iter_collected = cache.iter().collect::<Vec<_>>();
     let zip_iter_collected = cache.keys()
         .zip(cache.values())
@@ -443,6 +537,17 @@ fn separated_iters_zip_to_pair_iter() {
     assert_eq!(pair_iter_collected, zip_iter_collected);
 }
 
+#[test]
+fn separated_into_iters_zip_to_pair_into_iter() {
+    let cache = large_test_cache();
+    let pair_iter_collected = cache.clone().into_iter().collect::<Vec<_>>();
+    let zip_iter_collected = cache.clone().into_keys()
+        .zip(cache.into_values())
+        .collect::<Vec<_>>();
+
+    assert_eq!(pair_iter_collected, zip_iter_collected);
+}
+
 #[test]
 fn drain_clears_cache() {
     let mut cache = LruCache::new(1024);
@@ -455,24 +560,31 @@ fn drain_clears_cache() {
     assert!(cache.get("hello").is_none());
 }
 
+fn test_owning_iterator<I>(mut iter: I)
+where
+    I: Iterator<Item = (&'static str, &'static str)> + DoubleEndedIterator
+{
+    assert_eq!(Some(("hello", "world")), iter.next());
+    assert_eq!(Some(("good morning", "jupiter")), iter.next_back());
+    assert_eq!(Some(("hi", "venus")), iter.next_back());
+    assert_eq!(Some(("ahoy", "mars")), iter.next_back());
+    assert_eq!(Some(("greetings", "moon")), iter.next());
+    assert_eq!(None, iter.next_back());
+    assert_eq!(None, iter.next());
+}
+
 #[test]
 fn drain_returns_entries() {
-    let mut cache = LruCache::new(1024);
-    cache.insert("hello", "world").unwrap();
-    cache.insert("greetings", "moon").unwrap();
-    cache.insert("ahoy", "mars").unwrap();
-    cache.insert("hi", "venus").unwrap();
-    cache.insert("good morning", "jupiter").unwrap();
+    let mut cache = large_test_cache();
+    let drain = cache.drain();
+    test_owning_iterator(drain);
+}
 
-    let mut drain = cache.drain();
-
-    assert_eq!(Some(("hello", "world")), drain.next());
-    assert_eq!(Some(("good morning", "jupiter")), drain.next_back());
-    assert_eq!(Some(("hi", "venus")), drain.next_back());
-    assert_eq!(Some(("ahoy", "mars")), drain.next_back());
-    assert_eq!(Some(("greetings", "moon")), drain.next());
-    assert_eq!(None, drain.next_back());
-    assert_eq!(None, drain.next());
+#[test]
+fn into_iter_returns_entries() {
+    let cache = large_test_cache();
+    let into_iter = cache.into_iter();
+    test_owning_iterator(into_iter);
 }
 
 #[test]
@@ -510,3 +622,38 @@ fn clone_creates_independent_cache() {
     assert_eq!(cache_drained_expected, cache_drained);
     assert_eq!(clone_drained_expected, clone_drained);
 }
+
+fn assert_is_empty<T, I>(mut iterator: I)
+where
+    I: Iterator<Item = T> + DoubleEndedIterator
+{
+    assert!(iterator.next().is_none());
+    assert!(iterator.next_back().is_none());
+}
+
+#[test]
+fn empty_cache_builds_empty_iterators() {
+    let mut cache: LruCache<String, String> = LruCache::new(1024);
+
+    assert_is_empty(cache.iter());
+    assert_is_empty(cache.keys());
+    assert_is_empty(cache.values());
+    assert_is_empty(cache.drain());
+    assert_is_empty(cache.clone().into_iter());
+    assert_is_empty(cache.clone().into_keys());
+    assert_is_empty(cache.into_values());
+}
+
+#[test]
+fn touching_in_singleton_works() {
+    // Note: This weirdly specific test case isolates a previous bug.
+
+    let mut cache = LruCache::new(1024);
+    cache.insert("hello", "world").unwrap();
+    cache.touch("hello");
+
+    let mut iter = cache.keys();
+
+    assert_eq!(Some(&"hello"), iter.next());
+    assert_eq!(None, iter.next());
+}
